/*
 * PlayerSession.java
 */

package pps.server;

import java.net.*;
import java.io.*;
import java.util.LinkedList;
import pps.model.Vector2D;

/**
 * Class that acts as the player session for the connected client. Each playerhas 
 * a player session. All messages that are sent and received to/from the player goes 
 * through the player session. The player session passes incomming messages to the server, 
 * which in turn handles them properly.
 *
 * @author ivind
 */
public class PlayerSession{
    
	//Variables::
    private Socket socket = null;
    private SessionReceiver receiver;
    private SessionSender sender;
    private ServerConnector server;
    private PrintWriter out;
    private BufferedReader in;
   
    //Player variables::
    private int playerID;//The ID of the player
    private Vector2D playerPos;
    private String ipAdress;
    
    /** Creates a new instance of PlayerSession */
    public PlayerSession(Socket socket,ServerConnector server,int id,Vector2D pos){        
	this.socket = socket;
        receiver=new SessionReceiver(this);
        sender=new SessionSender();
        new Thread(receiver).start();
        new Thread(sender).start();        
        playerID=id;
        playerPos=pos;
        ipAdress=socket.getInetAddress().toString();
        
        //Tries to set up the in-going and out.going network streams::
        try{
            out=new PrintWriter(this.socket.getOutputStream(), true);
            //System.out.println("Client "+playerID+"# connected..");
            in = new BufferedReader(new InputStreamReader(this.socket.getInputStream()));    
        }catch(Exception e){
            e.printStackTrace();
        }
        this.server=server;                
    }
    
    /**
     * Return the position of the player assiciated with this session
     */
    public Vector2D getPlayerPos() {
		return playerPos;
	}
    
    /**
     * Sets the position of the player associated with this session
     */
	public void setPlayerPos(Vector2D playerPos) {
		this.playerPos = playerPos;
	}
    
    /**
     * Returns the player object of this class
     */
    public int getPlayerID(){
        return playerID;
    }
    
    /**
     * Set the player object of this class
     */
    public void setPlayerID(int p){
        playerID=p;
    }
    
    /**
     * Returns the ip address of the player 
     */
    public String getIpAdress() {
		return ipAdress;
	}
    
    /**
     * Sends the mesasge to the client.
     */
    public void sendMessage(String message){
        sender.sendMessage(message);
    }
    
    /**
     * Private method that closes all I/O streams, and stops the receiver and sender threads
     */
    private void close() {
		sender.stopSending();
	        receiver.stopReceiving();
	        try {	       
		    out.close();
		    in.close();
		    socket.close();
		} catch (Exception e) {
		    e.printStackTrace();
		}
    }
    
    /**
     * Closes all streams and ends the receiver and sender threads.
     */
    public void disconnect(){        
        close();
    }
    
    /**
     * Private class that receives incoming messages and sends them to the server
     */
    private class SessionReceiver implements Runnable{
        
    	//Variables::
        private boolean running=true;
        private PlayerSession parent;
        
        //Constructor::
        public SessionReceiver(PlayerSession p){
            parent=p;
        }
        
        /*
         * Starts the message receiving loop which reads messages and sends them to the mesasge
         * handler.
         */
        public void run() {
            while(socket==null){/*wait*/}
            while(running){
                try{
                    String message=in.readLine();
                    String m2=message.substring(0, message.length()-1);
                    char c; 
                    //System.out.println("S:"+playerID+"# ,M: "+message+" ,M2:"+m2);
                    server.handleMessage(m2,parent);                    
                }catch(Exception e){
                    //e.printStackTrace();
                }
            }            
        }
        
        /*
         * Sets the running boolean false, stopping the receiver loop
         */
        public void stopReceiving(){
            running=false;
        }    
    }
    /**
     * Private class that sends messages to the client connection. Outgoing mesages are
     * put into a mesage queue where the class handles them one at a time, and waits
     * for new messages when the queue is emptied.
     */
    private class SessionSender implements Runnable{
    	
    	//Variable::
        private LinkedList<String> messageQueue;
        boolean running=true;
        
        //Constructor::
        public SessionSender(){
            messageQueue=new LinkedList<String>();
        }
        
        /*
         * Sends the message to the player assiciated with the session
         */
        public void sendMessage(String message){
            messageQueue.add(message);
            synchronized(this){
                this.notify();
            }
        }   
        
        /*
         * Sets the running boolean to false, stopping the sending loop
         */
        public void stopSending(){
            running=false;
        } 
        
        /*
         * The loop that sends messages in the message queue to the player accosiated with
         * this session. The loop waits for outgoing mesasges when the message queue is empty,
         * and will send all of the messages in the message queue until the queue is empty
         * after having being notified that new messages have entered the queue.
         */
        public void run() {
            while(running){
                synchronized(this){
                    try{
                        while(!messageQueue.isEmpty()){
                            String m=messageQueue.getFirst();
                            out.println(m);
                            messageQueue.removeFirst();
                            //System.out.println("Sent message: "+m);
                            Thread.sleep(5);
                        }
                        this.wait();
                    }catch(Exception e){
                        e.printStackTrace();
                    }
                }
            }
        }
    } 
}

