/*
 * ServerConnector.java
 */

package pps.server;

import java.net.*;
import java.io.*;
import java.util.ArrayList;
import pps.model.LevelCoordinator;
import pps.model.Vector2D;

/**
 * The main server that listens to incomming connections, creates player sessions, and handles
 * incomming messages
 *
 * @author ivind
 */
public class ServerConnector implements Runnable,MessageListener{
    
    //Connection list
    private ArrayList<PlayerSession> playerList=new ArrayList<PlayerSession>();
    
    //Variables::
    private MessageHandler mHandler;
    private LevelCoordinator lvlHandler;
    private int position_interval=1000;
    
    //Constants::
    private final int SERVER_PORT=5338;
    
    /** Creates a new instance of ServerConnector */
    public ServerConnector(Integer i) {      
        mHandler=new MessageHandler(this);
        lvlHandler=new LevelCoordinator();
        if(i!=null)position_interval=i.intValue();
        
        //Send a position request every # seconds::
        try{
            new Thread(
                new Runnable(){
                    public void run(){
                        try{
                            while(true){
                                broadcastMessage(mHandler.createPositionRequest());
                                Thread.sleep(position_interval);
                            }
                        }catch(Exception e){e.printStackTrace();}
                    }
                }
            ).start();
        }catch (Exception e){}
    }
    
    /**
     * Adds the session to the player session lsit
     */
    public void addPlayerSession(PlayerSession session){
        playerList.add(session);
    }
    
    /**
     * Removes the player sessions from the list of player sessions
     */
    public void removePlayerSession(PlayerSession ps){
        ps.disconnect();
        playerList.remove(ps);
    }
        
    /**
     * Handles incoming messages from the clients
     */
    public void handleMessage(String message,PlayerSession p){
        mHandler.handleMessage(message,p);
    }
    
    /**
     * Broadcasts the message to all clients
     */
    public void broadcastMessage(String message){
        for(int i=0;i<playerList.size();i++){
            PlayerSession pl=playerList.get(i);
            if(pl==null){
                playerList.remove(i);
            }else{
                pl.sendMessage(message);
            }
        }
    }
    
    /**
     * Broadcast the mesage to all clients except the client 'ps'
     */
    public void broadcastMessageExcept(String message,PlayerSession ps){
    	for(int i=0;i<playerList.size();i++){
            PlayerSession pl=playerList.get(i);
            if(pl==null){
                playerList.remove(i);
            }else{
            	//Check to see whether the client is the paramter
            	if(ps.getPlayerID()!=pl.getPlayerID()){
            		pl.sendMessage(message);
                }
            }
        }
    }
    
    /**
     * Sends a message to the player session with the parameter 'id'
     */
    public void sendMessageToPLayerSession(int id,String message){
    	for(int i=0;i<playerList.size();i++){
    		PlayerSession pl=playerList.get(i);
    		//Checks to see whether the parameter id is the same as the player id
    		if(pl.getPlayerID()==id){
    			pl.sendMessage(message);
    			return;
    		}
    	}
    }
    
    /**
     * Checks whether a ip adress is already connected to the server
     */
    public boolean IPExist(String ip){
    	//Goes through the player list checking the ip of each player:: 
    	for(int i=0;i<playerList.size();i++){
    		PlayerSession ps=playerList.get(i);
    		
    		//Check the player IP versus the paramter IP::
    		if(ps.getIpAdress().equals(ip)){
    			return true;
    		}
    	}
    	//No player has the ip, return false::
    	return false;
    }
    
    /**
     * Server loop that listens to incoming connections. Creates a player session for each 
     * new incomming connection
     */
    public void run(){
        ServerSocket serverSocket = null;
        boolean listening = true;
        //Main loop listeneing for incoming connections::
        try {
            serverSocket = new ServerSocket(SERVER_PORT);
            System.out.println("<::Server started::>");
            int i=0;
            while (listening){
                //Accept incoming connections and add them to the player session list::
                PlayerSession session=new PlayerSession(serverSocket.accept(),this,i,new Vector2D(0,0));                
                addPlayerSession(session);
                i++;
            }
            serverSocket.close();
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(-1);
        } 
    }
    
      //**************************************************************************//
     //***                   Message handler methods                          ***//
    //**************************************************************************//
    
    /**
     * A client has sent a message saying that it wants to club another player. Send the message
     * to the client in question. 
     */
    public void notifyClubRequest(int id,boolean right, PlayerSession ps) {
    	sendMessageToPLayerSession(id, mHandler.createInitClubbing(right));
	}
    
	public void notifyCreatePlayer(int id, Vector2D pos, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	
	/**
	 * The server has a message from a client saying that it wants to disconnect, so the
	 * server should close the client stream, and remove the player session as well as 
	 * notify the rest of the clients that a player has disconnected
	 */
	public void notifyDisconnectRequest(PlayerSession ps) {
		broadcastMessageExcept(mHandler.createPlayerDisconnect(ps.getPlayerID()),ps);
		removePlayerSession(ps);
	}
	
	/**
	 * A client has sent a message that it has reached the exit of a level, so the server should 
	 * load a new level and tell all of the clients to load the new level.
	 */
	public void notifyExitReached(PlayerSession ps) {
		lvlHandler.nextLevel();
		broadcastMessage(mHandler.createLoadLevel(lvlHandler.getCurrentLevel(), lvlHandler.getStartingPoint()));
	}
	
	public void notifyInitClubbing(boolean right, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	public void notifyJoinAck(int id, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	
	/**
	 * The client has sent a message that it wants to join, so send a message back with the 
	 * appripriate client-connection protocoll
	 */
	public void notifyJoinRequest(PlayerSession ps) {		
		String m=mHandler.createJoinAck(ps.getPlayerID());
		ps.sendMessage(m);
		m="";
		
		//Add the current level::
		m+=mHandler.createLoadLevel(lvlHandler.getCurrentLevel(),lvlHandler.getStartingPoint());
		
		//Add players already connected to the server::
		for(int i=0;i<playerList.size();i++){
			PlayerSession p=playerList.get(i);
			if(ps.getPlayerID()!=p.getPlayerID()){
				m+=mHandler.createCreatePlayer(p.getPlayerID(), p.getPlayerPos());
			}
		}
		
		//Send the message::
		ps.sendMessage(m);	
		
		//Broadcast the new player to the other players::
		broadcastMessageExcept(mHandler.createCreatePlayer(ps.getPlayerID(), ps.getPlayerPos()), ps);
	}
	
	public void notifyLoadLevel(String lvlName,Vector2D startPos, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	public void notifyPlayerDisconnect(int id, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	
	/**
	 * The client has sent a message containing its position, and the other players should be 
	 * notified about the new position of the player
	 */
	public void notifyPosition(Vector2D pos, PlayerSession ps) {
		//Checks to see whether the position of the player has changed::
		if(pos.x!=ps.getPlayerPos().x || pos.y!=ps.getPlayerPos().y){
			ps.setPlayerPos(pos);
			broadcastMessageExcept(mHandler.createPlayerPositionChange(ps.getPlayerID(), pos) , ps);
		}
	}
	
	public void notifyPositionChange(int id, Vector2D pos, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	public void notifyPositionRequest(PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	public void notifyTriggerEvent(int eventID, PlayerSession ps) {
		// TODO Auto-generated method stub		
	}
	
	/**
	 * A client has sent a message saying that a event should be triggered so the server
	 * should notify all of the clients to trigger the event locally.
	 */
	public void notifyTriggerEventRequest(int eventID, PlayerSession ps) {
		broadcastMessage(mHandler.createTriggerEvent(eventID));		
	}	
	
    /*
     * Starts the server thread
     */
    public static void main(String arg[]){
    	Integer i=null;
    	try{
    		i=Integer.parseInt(arg[0]);
    	}catch(Exception e){
    		
    	}
        ServerConnector server=new ServerConnector(i);
        new Thread(server).start();        
    }
}
